//
//  lcd7seg.h
//
//  7-Sengent LCD Control
//  Copyright (C) 2011 Toyohiko Togashi tog001@nifty.com
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//	Description
//		7-Segment LCD control subroutine.
//		Supported any digits and order.
//
//  Device: PIC16F1939       IDE: MPLAB IDE 8.63, HI-TECH C Tools for the PIC10/12/16 MCU Family Version 9.81
//
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2011.03.16 v0.0  1st rel
//

#include "htc.h"						// HI-TECH C
#include "lcd7seg.h"					// Private

volatile unsigned char	LCDDATA[12] @ 0x7A0;

// LCD segment mapping
LCD7SEG_SEGMAP;							// See lcd7seg.h

// 7-Segment font
const  unsigned char lcd7seg_Pat[16] = {
//		  .gfedcba
		0b00111111,		// 0
		0b00000110,		// 1	 aaa
		0b01011011,		// 2	f   b
		0b01001111,		// 3	 ggg
		0b01100110,		// 4	e   c
		0b01101101,		// 5	 ddd .
		0b01111101,		// 6
		0b00000111,		// 7
		0b01111111,		// 8
		0b01101111,		// 9
		0b01110111,		// A
		0b01111100,		// B
		0b00111001,		// C
		0b01011110,		// D
		0b01111001,		// E
		0b01110001		// F
};

void lcd7seg_PutPattern(unsigned char pat, unsigned char pos){
	unsigned char	i;
	unsigned char	e;

	if (pos == 0) {						// All column
		e = LCD7SEG_DIG_WIDTH;
	} else if (pos <= LCD7SEG_DIG_WIDTH){
		e = pos;
		--pos;
	} else {							// Out of range
		e = 0;
	}
	for(; pos < e; pos++) {
		for(i = 0; i < 7; i++) {		// Set a-g segment
			unsigned char	m;
			unsigned char	f;

			m = lcd7seg_segmap[pos][i];
			f = 1 << (m & 0x07);
			if (pat & 1) {
				LCDDATA[m >> 3] |= f;
			} else {
				LCDDATA[m >> 3] &= ~f;
			}
			pat >>= 1;
		}
	}
	return;
}

void lcd7seg_DP(unsigned char flag, unsigned char pos){
	unsigned char	i;
	unsigned char	e;

	if (pos == 0) {						// All column
		e = LCD7SEG_DIG_WIDTH;
	} else if (pos <= LCD7SEG_DIG_WIDTH){
		e = pos;
		--pos;
	} else {							// Out of range
		e = 0;
	}
	for(; pos < e; pos++) {
		unsigned char	m;
		unsigned char	f;

		m = lcd7seg_segmap[pos][7];
		f = 1 << (m & 0x07);
		if (flag) {
			LCDDATA[m >> 3] |= f;
		} else {
			LCDDATA[m >> 3] &= ~f;
		}
	}
	return;
}

void led7seg_Dim(signed char dim){
	LCDCST = dim;
	if (dim > 0) {
		LCDCST = 0;						// Max
	} else if (dim == 0) {
		LCDCST = 7;						// Pale
	} else {
		LCDCST = ~dim & 0x07;
	}
}

void lcd7seg_DispInt(int value, unsigned char pos, signed char len) {
	unsigned char s;
	unsigned char z;

	if (value < 0) {
		value = -value;
		s = 1;
	} else {
		s = 0;
	}

	if (len < 0) {
		len = -len;
		z = 0b00111111;					// "0"
	} else {
		z = 0b00000000;					// " "
	}

	len--;
	do {
		unsigned char	c;

		c = value % 10;
		value /= 10;
		lcd7seg_PutHex(c, pos + len);
		--len;
	} while((len >= 0) && (value != 0));

	if (len >= 0) {
		if (s) {
			lcd7seg_PutPattern(0x40, pos + len);		// minus pattern
			--len;
		}
		for( ; len >= 0; --len) {
			lcd7seg_PutPattern(z, pos + len);			// zero suppless
		}
	}

	return;
}

void lcd7seg_Start(void){
	unsigned char	i;

	// Enable Internal LCD driver

	// Step.1
	LCDPSbits.LP = 0b0011;
	// Step.2
	LCDSE0 = (((1UL << LCD7SEG_SEG_PIN) - 1) >> 0 ) & 0xff;
	LCDSE1 = (((1UL << LCD7SEG_SEG_PIN) - 1) >> 8 ) & 0xff;
	LCDSE2 = (((1UL << LCD7SEG_SEG_PIN) - 1) >> 16) & 0xff;
	// Step.3
	LCDCONbits.LMUX  = LCD7SEG_COM_PIN - 1;
	LCDCONbits.CS    = 0b10;			// LFINTOSC
	LCDCONbits.SLPEN = 0;				// Active
	// Step.4
	for(i = 0; i < sizeof(LCDDATA); i++) {
		LCDDATA[i]  = 0b00000000;
	}
	// Step.5
	PIR2bits.LCDIF = 0;
	// Step.6
	LCDRLbits.LRLAP = 0b10;				// Power mode A: Mid power
	LCDRLbits.LRLBP = 0b01;				// Power mode B: Low power
	LCDRLbits.LRLAT = 0b001;			// Power mode A time: 1/16clock(WFT=0)
	LCDREFbits.LCDIRE = 1;
	LCDREFbits.LCDIRS = 0;
//	LCDREFbits.LCDIRI = 1;
	// Step.7
	LCDCONbits.LCDEN = 1;
	// Others
//	LCDPSbits.WFT    = 1;
//	LCDPSbits.BIASMD = 1;
//	LCDCST = 0;

#ifdef LCD7SEG_INIT_PATT
	lcd7seg_PutPattern(LCD7SEG_INIT_PATT, 0);
#endif
}

void lcd7seg_Stop(void){
	LCDCON = 0;
}

