//
//  main.c
//
//  Watt Hour Meter 0.7 / Battery operation version
//  Copyright(C)2011 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//	Description
//		Simple Watt hour meter.
//		4 channel CT(Current transformer) or
//		3 channel CT and voltage detect.
//		About one year battery operation.
//
//	Usage
//		Install CT and or voltage line clip to distribution board.
//		Insert AAA x 3 battery.
//		Push-switch 1+2, input configuration parameter and date time.
//		Push-switch 1, Display item
//		Push-switch 2, Channel+
//		Push-switch 3, Channel-
//		Push-switch 2+3, Auto channel display mode
//		Push-switch 1+3, History display mode
//
//  Device: PIC16F1939
//
//  Pin assign:
//		p1:  RE3/VPP/MCLR                               SW1 
//		p2:  RA0/SEG12/VCAP/SS/SRNQ/C2OUT/C12IN0-/AN0   LCD-21
//		p3:  RA1/SEG7/C12IN1-/AN1                       LCD-20
//		p4:  RA2/COM2/DACOUT/VREF-/C2IN+/AN2            LCD-28 
//		p5:  RA3/SEG15/VREF+/C1IN+/AN3                  LCD-19
//		p6:  RA4/SEG4/SRQ/T0CKI/CPS6/C1OUT              LCD-18
//		p7:  RA5/SEG5/VCAP/SS/SRNQ/CPS7/C2OUT/AN4       LCD-17
//		p8:  RE0/SEG21/CCP3/P3A/AN5                     LCD-16
//		p9:  RE1/SEG22/P3B/AN6                          LCD-15
//		p10: RE2/SEG23/CCP5/AN7                         CH4
//		p11: VDD                                        +4.5V
//		p12: VSS                                        GND
//		p13: RA7/SEG2/CLKIN/OSC1                        LCD-12
//		p14: RA6/SEG1/VCAP/CLKOUT/OSC2                  LCD-10
//		p15: RC0/P2B/T1CKI/T1OSO                        Crystal 32.768kHz
//		p16: RC1/P2A/CCP2/T1OSI                         Crystal 32.768kHz
//		p17: RC2/SEG3/P1A/CCP1                          LCD-6
//		p18: RC3/SEG6/SCK/SCL                           LCD-4
//		p19: RD0/COM3/CPS8                              SW2
//		p20: RD1/CCP4/CPS9                              SW3
//		p21: RD2/CPS10/P2B                              -
//		p22: RD3/CPS11/P2C/SEG16                        LCD-1
//		p23: RC4/SDI/SDA/T1G/SEG11                      LCD-2
//		p24: RC5/SDO/SEG10                              LCD-3
//		p25: RC6/TX/CK/SEG9                             LCD-5
//		p26: RC7/RX/DT/SEG8                             LCD-7
//		p27: RD4/CPS12/P2D/SEG17                        LCD-8
//		p28: RD5/CPS13/P1B/SEG18                        LCD-9
//		p29: RD6/CPS14/P1C/SEG19                        LCD-11
//		p30: RD7/CPS15/P1D/SEG20                        LCD-13
//		p31: VSS                                        GND
//		p32: VDD                                        +4.5V
//		p33: RB0/AN12/CPS0/SRI/INT/SEG0                 LCD-14
//		p34: RB1/AN10/C12IN3-/CPS1/VLCD1                CH1
//		p35: RB2/AN8/CPS2/VLCD2                         CH2
//		p36: RB3/AN9/C12IN2-/CPS3/CCP2/P2A/VLCD3        CH3
//		p37: RB4/AN11/CPS4/COM0                         LCD-29
//		p38: RB5/AN13/CPS5/CCP3/P3A/T1G/COM1            LCD-30
//		p39: RB6/ICSPCLK/ICDCLK/SEG14                   LCD-22
//		p40: RB7/ICSPDAT/ICDDAT/SEG13                   LCD-23
//
//	Clock:	Fosc=16MHz
//
//	Internal peripheral modules:
//		BOR:	-
//		Timer0:	-
//		Timer1:	1Hz interval(32.768kHz osc)
//		Timer1:	2Hz interval(32.768kHz osc)
//		Timer2:	-
//		Timer4:	-
//		Timer6:	-
//		Comp:	-
//		CCP4:	-
//		CCP5:	-
//		ECCP1:	-
//		ECCP2:	-
//		ECCP3:	-
//		MSSP:	-
//		EUSART:	-
//		ADC:	Measuering Current and voltage
//		SRLatch:-
//		LCD:	7-Segment panel drive
//
//	Environment:
//		IDE: MPLAB IDE 8.36
//      C Compiler for PIC18 MCUs v3.33
//
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2011.04.07 v0.7  First release
//    2011.04.25 v0.7a Power tune
//    2011.06.01 v0.7b Power tune 2, Scratch and Rebuild
//    2011.06.11 v0.7b Auto range x1,x2,x4
//    2011.07.03 v0.7b Support peak power and daily watt hour
//

#include <math.h>						// libc
#include <string.h>						// libc
#include "htc.h"						// HI-TECH C
#include "lcd7seg.h"					// Private


//*********************************************************************
//
//	Definition
//
//*********************************************************************

//
//	CPU Configuration Fuses bits
//
__CONFIG(
//	FOSC_ECH &		// ECH, External Clock, High Power Mode (4-32 MHz): device clock supplied to CLKIN pin
//	FOSC_ECM & 		// ECM, External Clock, Medium Power Mode (0.5-4 MHz): device clock supplied to CLKIN pin
//	FOSC_ECL &		// ECL, External Clock, Low Power Mode (0-0.5 MHz): device clock supplied to CLKIN pin
	FOSC_INTOSC &	// INTOSC oscillator: I/O function on CLKIN pin
//	FOSC_EXTRC &	// EXTRC oscillator: External RC circuit connected to CLKIN pin
//	FOSC_HS &		// HS Oscillator, High-speed crystal/resonator connected between OSC1 and OSC2 pins
//	FOSC_XT &		// XT Oscillator, Crystal/resonator connected between OSC1 and OSC2 pins
//	FOSC_LP &		// LP Oscillator, Low-power crystal connected between OSC1 and OSC2 pins
//	WDTE_ON &		// WDT enabled
//	WDTE_NSLEEP &	// WDT enabled while running and disabled in Sleep
//	WDTE_SWDTEN &	// WDT controlled by the SWDTEN bit in the WDTCON register
	WDTE_OFF &		// WDT disabled
	PWRTE_OFF &		// PWRT disabled
//	PWRTE_ON &		// PWRT enabled
//	MCLRE_ON &		// MCLR/VPP pin function is MCLR
	MCLRE_OFF &		// MCLR/VPP pin function is digital input
	CP_OFF &		// Program memory code protection is disabled
//	CP_ON &			// Program memory code protection is enabled
	CPD_OFF &		// Data memory code protection is disabled
//	CPD_ON &		// Data memory code protection is enabled
//	BOREN_ON &		// Brown-out Reset enabled
//	BOREN_NSLEEP &	// Brown-out Reset enabled while running and disabled in Sleep
//	BOREN_SBODEN &	// Brown-out Reset controlled by the SBOREN bit in the BORCON register
	BOREN_OFF &		// Brown-out Reset disabled
	CLKOUTEN_OFF &	// CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin
//	CLKOUTEN_ON &	// CLKOUT function is enabled on the CLKOUT pin
//	IESO_ON &		// Internal/External Switchover mode is enabled
	IESO_OFF &		// Internal/External Switchover mode is disabled
//	FCMEN_ON &		// Fail-Safe Clock Monitor is enabled
	FCMEN_OFF &		// Fail-Safe Clock Monitor is disabled
	0x3fff);

__CONFIG(
	WRT_OFF &		// Write protection off
//	WRT_BOOT &		// 000h to 1FFh write protected, 200h to 3FFFh may be modified by EECON control
//	WRT_HALF &		// 000h to 1FFFh write protected, 2000h to 3FFFh may be modified by EECON control
//	WRT_ALL &		// 000h to 3FFFh write protected, no addresses may be modified by EECON control
	VCAPEN_OFF &	// All VCAP pin functionality is disabled
//	VCAPEN_RA6 &	// VCAP functionality is enabled on RA6
//	VCAPEN_RA5 &	// VCAP functionality is enabled on RA5
//	VCAPEN_RA0 &	// VCAP functionality is enabled on RA0
//	PLLEN_ON &		// 4x PLL enabled
	PLLEN_OFF &		// 4x PLL disabled
	STVREN_ON &		// Stack Overflow or Underflow will cause a Reset
//	STVREN_OFF &	// Stack Overflow or Underflow will not cause a Reset
	BORV_19 &		// Brown-out Reset Voltage (VBOR) set to 1.9 V
//	BORV_25 &		// Brown-out Reset Voltage (VBOR) set to 2.5 V
//	LVP_ON &		// Low-voltage programming enabled
	LVP_OFF &		// High-voltage on MCLR/VPP must be used for programming
	0x3fff);

//
// Application constant
//
//#define _XTAL_FREQ	16000000		// [Hz]	-> IDE parameter

#define CH_COUNT		4				// Current & voltage channel

#define ADC_VREF		1.024			// [V] adc_range==1
#define ADC_RESOLUTION	1024			// 10bit
#define ADC_BUF_SZ		80				// Capture buffer size [x2byte]
#define KWH_PRICE		25				// [YEN/kWh] default
#define LINE_FREQ		50				// [Hz] default
#define MAX_SAMPLE_COUNT	128			// 

#define CT_RL			330				// [ohm]
#define CT_N			3000			// U_RD CTL-10-CLS datasheet
#define CT_K			1.0				// U_RD CTL-10-CLS datasheet

#define IN_V			100				// [V] RMS, 100V fixed mode
#define R_DIV1			330				// [ohm] Voltage sense register, adc side
#define R_DIV2			200000			// [ohm] Voltage sense register, line side

#define SW1_PIN			!(PORTEbits.RE3)
#define SW2_PIN			!(PORTDbits.RD0)
#define SW3_PIN			!(PORTDbits.RD1)
#define SW1_INIT()		{TRISDbits.TRISD0 = 1; ANSELDbits.ANSD0 = 0; }
#define SW2_INIT()		{TRISDbits.TRISD1 = 1; ANSELDbits.ANSD1 = 0; }
#define SW3_INIT()		// {TRISEbits.TRISE3 = 1; ANSELEbits.ANSE3 = 0; }

// Inline function
#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define ABS(a)		(((a) >  0 ) ? (a) : -(a))


//
// Global area
//
const unsigned char	adc_anx[CH_COUNT] = {10, 8, 9, 7};		// ANn
const unsigned char	lcd_pat[CH_COUNT] = {0x40, 0x02, 0x01, 0x00};

unsigned char	ch_count  = CH_COUNT;	// 100V Fixed mode
unsigned char	kwh_price = KWH_PRICE;
unsigned char	line_freq = LINE_FREQ;
unsigned short	cycle_time = (32768 / LINE_FREQ + 32768);
unsigned char	adc_samp;				// Sampling count

short			adc_val[ADC_BUF_SZ][CH_COUNT];

short			v_val = IN_V;			// [0.1V]
short			a_val[CH_COUNT];		// [mA]
short			w_val[CH_COUNT];		// [W]
short			w_max[CH_COUNT] = {		// [W]
					0, 0, 0, 0			//
				};						//     
long			w_est[CH_COUNT] = {		// [Ws] - Watt/Second
					0, 0, 0, 0			//
				};						//     
long			w_day[CH_COUNT] = {		// [Ws] - Watt/Second
					0, 0, 0, 0			//
				};						//     

volatile int 	yy = 2011;				// Year
volatile char 	mm = 1;					// Month
volatile char 	dd = 1;					// Day of month
volatile char 	hh = 0;					// Hour
volatile char 	mi = 0;					// Minites
volatile char 	ss = 0;					// Second
char ddom[] = {							//
	0,									//
	31,28,31,30,31,30,31,31,30,31,30,31	//
};										//

unsigned short	ts_start;				// DEBUG
unsigned short	ts_measure;				// DEBUG
unsigned short	ts_calc;				// DEBUG
unsigned short	ts_disp;				// DEBUG
unsigned short	now_range;				// DEBUG


//*********************************************************************
//
//	Intrrupt handler
//
//*********************************************************************
//
// RTC  Time & Calender
//
void interrupt isr(void) {
	if (TMR1IE && TMR1IF) {

		TMR1 = -32768;					// 32.768Hz crystal
		TMR1IF = 0;
		ss++;							//
//		ss += 2;						// ADD 2011.04.25
		if (ss > 59) {
			ss = 0;
			mi++;
			if (mi > 59) {
				mi = 0;
				hh++;
				if (hh > 23) {
					hh = 0;
					dd++;
					if (((yy & 0x03) == 0) && ((yy % 100) != 0)) {
						ddom[2] = 29;
					} else {
						ddom[2] = 28;
					}
					if (dd > ddom[mm]) {
						dd = 1;
						mm++;
						if (mm > 12) {
							mm = 1;
							yy++;
						}
					}
				}
			}
		}
	}
	return;
}


//*********************************************************************
//
//	Utilitys
//
//*********************************************************************

//---------------------------------------------------------------------
//	A/D converter
//---------------------------------------------------------------------
void adc_start(char range) {
	// Setup chip pin
//	TRISAbits.TRISA0 = 1;		/* input */
//	ANSELAbits.ANSA0 = 1;		/* ANALOG */
//	TRISAbits.TRISA1 = 1;
//	ANSELAbits.ANSA1 = 1;
//	TRISAbits.TRISA2 = 1;
//	ANSELAbits.ANSA2 = 1;
//	TRISAbits.TRISA3 = 1;
//	ANSELAbits.ANSA3 = 1;
//	TRISAbits.TRISA5 = 1;
//	ANSELAbits.ANSA4 = 1;
//	TRISBbits.TRISB0 = 1;
//	ANSELBbits.ANSB0 = 1;
	TRISBbits.TRISB1 = 1;
	ANSELBbits.ANSB1 = 1;
	TRISBbits.TRISB2 = 1;
	ANSELBbits.ANSB2 = 1;
	TRISBbits.TRISB3 = 1;
	ANSELBbits.ANSB3 = 1;
//	TRISBbits.TRISB4 = 1;
//	ANSELBbits.ANSB4 = 1;
//	TRISBbits.TRISB5 = 1;
//	ANSELBbits.ANSB5 = 1;
//	TRISEbits.TRISE0 = 1;
//	ANSELEbits.ANSE0 = 1;
//	TRISEbits.TRISE1 = 1;
//	ANSELEbits.ANSE1 = 1;
	TRISEbits.TRISE2 = 1;
	ANSELEbits.ANSE2 = 1;

//	A/D module
	ADCON0bits.ADON   = 1;				// On
	ADCON1bits.ADPREF = 3;				// Vref+ = FVR
	ADCON1bits.ADNREF = 0;				// Vref- = Vss
	ADCON1bits.ADFM   = 1;				// Right jusstified
	ADCON1bits.ADCS   = 4;				// 4MHz Fosc/4 TAD=1us
//	ADCON1bits.ADCS   = 5;				// 16MHz Fosc/16 TAD=1us
//	ADCON1bits.ADCS   = 3;				// Frc TAD=1-6us

// FVR module
	FVRCONbits.FVREN = 1;				// On
//	FVRCONbits.ADFVR = 1;				// 1.024V
//	FVRCONbits.ADFVR = 2;				// 2.048V
//	FVRCONbits.ADFVR = 3;				// 4.096V
	FVRCONbits.ADFVR = range;
	while(!FVRCONbits.FVRRDY);			// PIC16F193x is never ready
	return;
}

void adc_stop() {
//	A/D module
	ADCON0 = 0;							// Disable
	ADCON1 = 0;
// FVR module
	FVRCON = 0;							// Disable
	return;
}

short adc_read(unsigned char channel) {
	ADCON0bits.CHS = channel;
//	__delay_us(9);						// Acquisition Time
	__delay_us(4);						// Acquisition Time
	ADCON0bits.ADGO = 1;
	while(ADCON0bits.ADGO);				// Conversion Time about 12us
//	return(ADRES);
	return((ADRES>1) ? ADRES : 0);
}

//---------------------------------------------------------------------
//	EEPROM
//---------------------------------------------------------------------
void eeprom_writes(unsigned char addr, unsigned char *buf, unsigned char len){
	for(; len > 0; --len) {
//		EEPROM_WRITE(addr++, *buf++);
		eeprom_write(addr++, *buf++);
	}
	return;
}
void eeprom_reads(unsigned char addr, unsigned char *buf, unsigned char len){
	for(; len > 0; --len) {
//		*buf++ = EEPROM_READ(addr++);	// Bug fix
		*buf++ = eeprom_read(addr++);
	}
	return;
}

#define EE_ENTRY_NUM	13
//#define EE_ENTRY_NUM	12				// DEBUG
struct ee_head {
	unsigned char	ch_count;
	unsigned char	line_freq;
	unsigned char	kwh_price;
	int				yy;
	signed char		mm;
	signed char		dd;
//unsigned short	ts_start;			// DEBUG
//unsigned short	ts_measure;			// DEBUG
//unsigned short	ts_calc;			// DEBUG
//unsigned short	ts_disp;			// DEBUG
//unsigned char	samp_ct;				// DEBUG
};
struct ee_entry {
	int			yy;
	signed char	mm;
	long		ws[CH_COUNT];
};

void ee_load(void){
	struct ee_head	ee;
	struct ee_entry	ent;
	unsigned char	i;

	eeprom_reads(0, (unsigned char *)&ee, sizeof(ee));
	if (ee.yy != 0xffff) {
		yy       = ee.yy;
		mm       = ee.mm;
		dd       = ee.dd;
		hh       = 0;
		mi       = 0;
		ss       = 0;
		ch_count  = ee.ch_count;
		line_freq = ee.line_freq;
		kwh_price = ee.kwh_price;
		for(i = 0; i < EE_ENTRY_NUM; i++) {
			eeprom_reads(i * sizeof(ent) + sizeof(ee), (unsigned char *)&ent, sizeof(ent));
			if ((ent.yy == yy) && (ent.mm == mm)) {
				break;
			}
		}
		if (i < EE_ENTRY_NUM) {
			memcpy(w_est, ent.ws, sizeof(w_est));
		} else {
			memset(w_est, 0, sizeof(w_est));
		}
	}
}

void ee_save(int yy, char mm, char dd){
	struct ee_head	ee;
	struct ee_entry	ent;
	unsigned char	i;
	int				min_yy;
	signed char		min_mm;
	unsigned char	min_pp;

	ee.yy        = yy;
	ee.mm        = mm;
	ee.dd        = dd;
	ee.ch_count  = ch_count;
	ee.line_freq = line_freq;
	ee.kwh_price = kwh_price;

//	ee.ts_start  = ts_start;			// DEBUG
//	ee.ts_measure = ts_measure;			// DEBUG
//	ee.ts_calc   = ts_calc;				// DEBUG
//	ee.ts_disp   = ts_disp;				// DEBUG
//	ee.samp_ct   = adc_samp;			// DEBUG

	eeprom_writes(0, (unsigned char *)&ee, sizeof(ee));

	min_yy = 2100;
	min_mm = 13;
	min_pp = 0;
	for(i = 0; i < EE_ENTRY_NUM; i++) {
		eeprom_reads(i * sizeof(ent) + sizeof(ee), (unsigned char *)&ent, sizeof(ent));
		if ((min_yy >= ent.yy) && (min_mm > ent.mm)) {
			min_yy = ent.yy;
			min_mm = ent.mm;
			min_pp = i;
		}
		if ((ent.yy == yy) && (ent.mm == mm)) {
			break;
		}
	}
	if (i >= EE_ENTRY_NUM) {
		i = min_pp;
	}
	ent.yy = yy;
	ent.mm = mm;
	memcpy(ent.ws, w_est, sizeof(ent.ws));
	eeprom_writes(i * sizeof(ent) + sizeof(ee), (unsigned char *)&ent, sizeof(ent));
}

//---------------------------------------------------------------------
//	Setup Date Time and parameters
//---------------------------------------------------------------------
void date_set(void) {
    enum { FREQ, CH, PRICE, YY, MM, DD, HH, MI, END } item;
	unsigned char	freq;
	unsigned char	ch;
	unsigned char	price;
	int				y;
	char			m;
	char			d;
	char			h;
	char			n;

	unsigned char	sw1;
	unsigned char	sw2;
	unsigned char	sw3;

	lcd7seg_DP(0, 0);					// Clear screen
	lcd7seg_PutPattern(0b00000000, 0);	// Clear screen
	while (SW1_PIN);	// wait key off
	while (SW2_PIN);
	while (SW3_PIN);

	freq	= line_freq;
	ch		= ch_count;
	price	= kwh_price;
	y   = yy;
	m   = mm;
	d   = dd;
	h   = hh;
	n   = mi;
	sw1 = 0;
	sw2 = 0;
	sw3 = 0;

	item = FREQ;
	for(;;) {
		// display
		if (TMR1H & 0x40) {				// Blink 1Hz
			switch(item){
			case FREQ:
				lcd7seg_PutPattern(0b00000000, 1);	//  
				lcd7seg_PutPattern(0b01110001, 2);	// F
				lcd7seg_PutPattern(0b01010000, 3);	// r
				lcd7seg_PutPattern(0b01100111, 4);	// q
				lcd7seg_DispInt(freq, 5, 4);
				break;
			case CH:
				lcd7seg_PutPattern(0b00000000, 1);	//  
				lcd7seg_PutPattern(0b00000000, 2);	//  
				lcd7seg_PutPattern(0b00111001, 3);	// C
				lcd7seg_PutPattern(0b01110100, 4);	// h
				lcd7seg_DispInt(ch, 5, 4);
				break;
			case PRICE:
				lcd7seg_PutPattern(0b00000000, 10);	// overlay column2
				lcd7seg_PutPattern(0b00000000, 1);	//  
				lcd7seg_PutPattern(0b01110011, 2);	// P
				lcd7seg_PutPattern(0b01010000, 3);	// r
				lcd7seg_PutPattern(0b01011000, 4);	// c
				lcd7seg_DispInt(price, 5, 4);
				break;
			case YY:
				lcd7seg_PutPattern(0b00000000, 1);	//  
				lcd7seg_PutPattern(0b00000000, 2);	//  
				lcd7seg_PutPattern(0b01101110, 3);	// y
				lcd7seg_PutPattern(0b01101110, 4);	// y
				lcd7seg_DispInt(y, 5, 4);
				break;
			default:
				lcd7seg_PutPattern(0x08, 9);
				lcd7seg_DispInt(m, 1, 2);
				lcd7seg_DispInt(d, 3, 2);
				lcd7seg_DispInt(h, 5, 2);
				lcd7seg_DispInt(n, 7, 2);
			}
		} else {
			switch(item) {
			case FREQ:
			case CH:
			case PRICE:
			case YY:
				lcd7seg_PutPattern(0x08, 5);
				lcd7seg_PutPattern(0x08, 6);
				lcd7seg_PutPattern(0x08, 7);
				lcd7seg_PutPattern(0x08, 8);
				break;
			case MM:
				lcd7seg_PutPattern(0x08, 1);
				lcd7seg_PutPattern(0x08, 2);
				break;
			case DD:
				lcd7seg_PutPattern(0x08, 3);
				lcd7seg_PutPattern(0x08, 4);
				break;
			case HH:
				lcd7seg_PutPattern(0x08, 5);
				lcd7seg_PutPattern(0x08, 6);
				break;
			default:
				lcd7seg_PutPattern(0x08, 7);
				lcd7seg_PutPattern(0x08, 8);
				break;
			}
		}
		
		// Check push
		if (SW1_PIN) {
			if (sw1 == 4) {		// Anti chattering
				item++;
				if (item == END) {	// Fix
					yy = y;
					mm = m;
					dd = d;
					hh = h;
					mi = n;
					ss = 0;
					line_freq = freq;
					cycle_time = 32768 / freq + 32768;
					ch_count = ch;
					kwh_price = price;
					break;		// forever exit
				}
			}
			sw1++;
			if (sw1 == 100) {		// long push - cancel
				break;
			}
		} else {
			sw1 = 0;
		}
		if (SW2_PIN) {
			if ((sw2 == 2) || (sw2 > 50)) {
				switch(item) {
				case FREQ:
					if (freq > 10) {
						--freq;
					}
					break;
				case CH:
					if (ch > 1) {
						--ch;
					}
					break;
				case PRICE:
					if (price > 1) {
						--price;
					}
					break;
				case YY:
					--y;
					if (y <= 0) {
						y = 9999;
					}
					break;
				case MM:
					--m;
					if (m == 0) {
						m = 12;
					}
					break;
				case DD:
					--d;
					if (d == 0) {
						d = ddom[m];
					}
					break;
				case HH:
					if (h <= 0) {
						h = 24;
					}
					--h;
					break;
				default:
					if (n <= 0) {
						n = 60;
					}
					--n;
					break;
				}
			}
			sw2++;
		} else {
			sw2 = 0;
		}
		if (SW3_PIN) {
			if ((sw3 == 2) || (sw3 > 50)) {
				switch(item) {
				case FREQ:
					if (freq < 100) {
						freq++;
					}
					break;
				case CH:
					if (ch < CH_COUNT) {
						ch++;
					}
					break;
				case PRICE:
					if (price < 100) {
						price++;
					}
					break;
				case YY:
					y++;
					if (y > 9999) {
						y = 1;
					}
					break;
				case MM:
					m++;
					if (m > 12) {
						m = 1;
					}
					break;
				case DD:
					d++;
					if (d > ddom[m]) {
						d = 1;
					}
					break;
				case HH:
					h++;
					if (h > 23) {
						h = 0;
					}
					break;
				default:
					n++;
					if (n > 59) {
						n = 0;
					}
					break;
				}
			}
			sw3++;
		} else {
			sw3 = 0;
		}

		__delay_ms(20);
	}
	
	lcd7seg_DP(0, 0);					// Clear screen
	lcd7seg_PutPattern(0b00000000, 0);	// Clear screen
	return;
}

//---------------------------------------------------------------------
//	History data
//---------------------------------------------------------------------
void hist_show(void) {
	unsigned char	sw1;
	unsigned char	sw2;
	unsigned char	sw3;
	unsigned char	i;
	struct ee_entry	ent;
	unsigned char	c;

	lcd7seg_DP(0, 0);					// Clear screen
	lcd7seg_PutPattern(0b00000000, 0);	// Clear screen
	while (SW1_PIN);	// wait key off
	while (SW2_PIN);
	while (SW3_PIN);

	c   = 0;
	i   = 0;
	sw1 = 0;
	sw2 = 0;
	sw3 = 0;
	lcd7seg_DP(0, 0);
	lcd7seg_DP(1, 7);

	for(;;) {

		// display
		eeprom_reads(i * sizeof(ent) + sizeof(struct ee_head), (unsigned char *)&ent, sizeof(ent));
		if (ent.yy == 0xffff) {
			memset(&ent, 0, sizeof(ent));
		}
		lcd7seg_PutPattern(lcd_pat[c], 9);	// Channel icon
		lcd7seg_DispInt(ent.yy, 1, 2);
		lcd7seg_DispInt(ent.mm, 3, 2);
		lcd7seg_DispInt((ent.ws[c]+180000) / 360000, 5, 4);		// 0.1kWh
		
		// Check push
		if (SW1_PIN) {
			if (sw1 == 4) {
				break;
			}
			sw1++;
		} else {
			sw1 = 0;
		}
		if (SW2_PIN) {
			if (sw2 == 2) {
				i++;
				if (i >= EE_ENTRY_NUM) {
					i = 0;
				}
			}
			sw2++;
		} else {
			sw2 = 0;
		}
		if (SW3_PIN) {
			if (sw3 == 2) {
				c++;
				if (c >= ch_count) {
					c = 0;
				}
			}
			sw3++;
		} else {
			sw3 = 0;
		}

		__delay_ms(20);
	}

	lcd7seg_DP(0, 0);					// Clear screen
	lcd7seg_PutPattern(0b00000000, 0);	// Clear screen
	return;
}

//*********************************************************************
//
//	M a i n
//
//*********************************************************************
main(void){
    enum { WH_W, YEN, VOL, AMP, PF, W_PEAK, DATE_TIME, END } disp_item;
    enum { SINGLE, AUTO } ch_mode;
    enum { MONTH, DAY } sum_period;
	unsigned char	disp_ch;
	int				yy_s;
	char			mm_s;
	char			dd_s;

	//
	// Initialization
	//

	// Main clock
	OSCCONbits.IRCF = 0b1111;			// Tosc=16MHz

	// Digital port
	SW1_INIT();
	SW2_INIT();
	SW3_INIT();

	// Timer1 on
	T1CONbits.T1OSCEN = 1;				// Low power OSC
	T1CONbits.TMR1CS  = 2;				// Crystal 32.768kHz
	T1CONbits.T1CKPS  = 0;				// 1/1
	T1CONbits.nT1SYNC = 1;				// Direct
	T1CONbits.TMR1ON  = 1;				// Start

	// LCD panel
	lcd7seg_Start();
	// Wakeup message
	lcd7seg_DP(0, 0);					// Clear screen
	lcd7seg_PutPattern(0b00000000, 0);	// Clear screen
	lcd7seg_PutPattern(0b00111110, 2);	// "W"
	lcd7seg_PutPattern(0b01110100, 3);	// "h"
	lcd7seg_PutPattern(0b00110111, 4);	// "m"
	lcd7seg_PutPattern(0b00001001, 10);	// ?
	lcd7seg_PutHex(0, 5);				// "0"
	lcd7seg_PutHex(7, 6);				// "7"
	lcd7seg_PutPattern(0b01111100, 7);	// "b"
	__delay_ms(1000);
	lcd7seg_PutPattern(0b00000000, 0);	// Clear screen

	// Application var
	disp_item = WH_W;
	ch_mode = AUTO;
	disp_ch   = 0;
	ee_load();
	yy_s = yy;
	mm_s = mm;
	dd_s = dd;
	sum_period = MONTH;

	// Intrrupt on
	PIE1bits.TMR1IE = 1;
	INTCONbits.PEIE = 1;
	INTCONbits.GIE  = 1;

	//
	// Forever loop
	//
	for(;;) {
		unsigned char	i;				// Sample counter
		unsigned char	j;				// Channel counter


		//-------------------------------------------------------------
		// Caputure one cycle wave
		//-------------------------------------------------------------
//		ts_start = TMR1;				// DEBUG
		{
			static char		range = 1;	// 1-3
			static short	bmax = 0;	// Before cycle maximum
			short			*p;

			// Range fix
			if (bmax > 1000) {
				if (range < 3) {
					range++;
				}
			} else if (bmax < 500) {
				if (range > 1) {
					--range;
				}
			}
			now_range = range;

			// A/D convert
			adc_start(range);
			p = adc_val[0];
			OSCCONbits.IRCF = 0b1101;	// Tosc=4MHz, slow down
			for(i = 0; i < ADC_BUF_SZ; i++) {
				unsigned char	b;
	
				// Wait sampling period
//				b = TMR1L & 0xf0;		// 488us
//				while((TMR1L & 0xf0) == b);
				b = TMR1L & 0xe0;		// 976us
				while((TMR1L & 0xe0) == b);

				// Scan analog port
				for(j = 0; j < CH_COUNT; j++) {
//					*p++ = adc_read(adc_anx[j]);
//					--------------Inline---------------
					ADCON0bits.CHS = adc_anx[j];
//					__delay_us(9);		// Acquisition Time
					__delay_us(4);		// Acquisition Time
					ADCON0bits.ADGO = 1;
					while(ADCON0bits.ADGO);	// Conversion Time about 12us
					*p++ = ADRES;
//					*p++ = (ADRES>1) ? ADRES : 0;
//					--------------Inline---------------
				}
				if (TMR1 > cycle_time) {
					break;
				}
			}
			OSCCONbits.IRCF = 0b1111;	// Tosc=16MHz, resume
			adc_stop();
			adc_samp = i + 1;

			// Range expand
			bmax = 0;
			p = adc_val[0];
			for(i = 0; i < adc_samp; i++) {
				for(j = 0; j < CH_COUNT; j++) {
					short	a;

					a = *p;
					if (bmax < a) {
						bmax = a;
					}
					switch(range) {
					case 2:
						a <<= 1;			// x2
						break;
					case 3:
						a <<= 2;			// x4
						break;
					default:
						break;
					}
					*p++ = a;
				}
			}
		}
//		ts_measure = TMR1;				// DEBUG


		//-------------------------------------------------------------
		// Calcurate, 100V fixed mode
		//-------------------------------------------------------------
		if (ch_count == CH_COUNT) {

			//
			// Calcurate
			//
			for(j = 0; j < ch_count; j++) {
				long	sum_a;
				short	v;
				short	a;
				short	w;

				sum_a = 0;
				for(i = 0; i < adc_samp; i++) {
					long	r;

					r = adc_val[i][j];
					sum_a += (r * r);	// RMS
				}

				v = IN_V;				// V
				a = sqrt((double)sum_a * 2.0 / (double)adc_samp)	// RMS
				  * (double)(ADC_VREF / ADC_RESOLUTION)
				  * (double)CT_N
				  / (double)(CT_K * CT_RL)
				  * 1000.0;				// mA
				w = (long)v_val			// V
				  * (long)a				// mA
				  / 1000L;				// W
	
				v_val    = v;
				a_val[j] = a;
				w_val[j] = w;
				w_est[j] += w;
			}


		//-------------------------------------------------------------
		// Calcurate, Real voltage mode
		//-------------------------------------------------------------
		} else {

			//
			// Create a mirror wave
			//
			{
				unsigned char	h;

				h = adc_samp / 2;			// 1cycle
				for(j = 0; j < CH_COUNT; j++) {
					for(i = 0; i < adc_samp; i++) {
						unsigned char	p;
		
						if (adc_val[i][j] < 1) {	//	Low level threshold
							p = i + h;
							if (p >= adc_samp) {
								p -= adc_samp;
							}
							adc_val[i][j] -= adc_val[p][j];
						}
					}
				}
			}
	
			//
			// Calcurate power and rms VA
			//
			{
				long			sum_v;
				long			sum_a[CH_COUNT];
				long			sum_w[CH_COUNT];
				unsigned char	f;		// Flag of RMS process
	
				// Initial
				f = ((disp_item == VOL) || (disp_item == AMP) || (disp_item == PF)) ? 1 : 0;
				sum_v = 0;
				for(j = 0; j < ch_count; j++) {
					sum_a[j] = 0;
					sum_w[j] = 0;
				}
	
				// Time slice
				for(i = 0; i < adc_samp; i++) {
					short	v;
					short	a;
	
					// Voltage
					v = adc_val[i][ch_count];
					if (f) {
						sum_v += ((long)v * (long)v);
					}
	
					for(j = 0; j < ch_count; j++) {
	
						// Current
						a = adc_val[i][j];
						if (f) {
							sum_a[j] += ((long)a * (long)a);
						}
	
						// Power
						sum_w[j] += (long)v * (long)a;

					}
				}
	
				// SUMMARY-AVERAGE
				if (f) {
					v_val = sqrt((double)sum_v / (double)adc_samp)				// RMS[V]
					      * (double)(ADC_VREF / ADC_RESOLUTION)
					      * (double)(R_DIV1 + R_DIV2)
					      / (double)(R_DIV1);
				}
				for(j = 0; j < ch_count; j++) {
					if (f) {
						a_val[j] = sqrt((double)sum_a[j] / (double)adc_samp)	// RMS[mA]
						         * (double)(ADC_VREF / ADC_RESOLUTION)
						         * (double)CT_N
						         / (double)(CT_K * CT_RL)
						         * 1000.0;
					}

					w_val[j] = (double)sum_w[j]		// Average[W]	***Self optimize***
					         / (double)adc_samp
					         * 0.0055187;

					w_est[j] += w_val[j];

					if (w_max[j] < w_val[j]) {
						w_max[j] = w_val[j];
					}
				}
			}
		}

		ts_calc = TMR1;


		//-------------------------------------------------------------
		// Regular process
		//-------------------------------------------------------------
		//
		// Daily task
		//
		if (dd_s != dd) {
			ee_save(yy_s, mm_s, dd_s);					// Check point

			for(j = 0; j < ch_count; j++) {
				w_max[j] = 0;
				w_day[j] = w_est[j];
			}

			dd_s = dd;
		}

		//
		// Monthly task
		//
		if (mm_s != mm) {
			memset(w_est, 0, sizeof(w_est));
			yy_s = yy;
			mm_s = mm;
		}

		//-------------------------------------------------------------
		// Input process
		//-------------------------------------------------------------
		//
		// Push switch event
		//
		if (SW1_PIN && SW2_PIN && SW3_PIN) {	// DEBUG
			int i;

			for(i = 0; i < 256; i += 2) {
				eeprom_writes(i, (unsigned char *)&adc_val[i >> 1][3], 2);
			}
		} else if (SW1_PIN && SW2_PIN) {
			date_set();
			ee_save(yy, mm, dd);
		} else if (SW1_PIN && SW3_PIN) {
			hist_show();
		} else if (SW2_PIN && SW3_PIN) {
			ch_mode = AUTO;
		} else if (SW1_PIN) {
			disp_item++;
			if (disp_item >= END) {
				disp_item = WH_W;
			}
			lcd7seg_DP(0, 0);					// Clear screen
			lcd7seg_PutPattern(0b00000000, 0);	// Clear screen
		} else if (SW2_PIN) {
			ch_mode = SINGLE;
			if (disp_ch == 0) {
				disp_ch = ch_count;
				sum_period = sum_period == MONTH ? DAY : MONTH;
			}
			--disp_ch;
		} else if (SW3_PIN) {
			ch_mode = SINGLE;
			disp_ch++;
			if (disp_ch >= ch_count) {
				disp_ch = 0;
				sum_period = sum_period == MONTH ? DAY : MONTH;
			}
		}

		//-------------------------------------------------------------
		// Output process
		//-------------------------------------------------------------
		//
		// Display
		//
		{
			unsigned char	c;
			long			e;
			short			y;

			if (ch_mode == AUTO) {
				disp_ch++;
				if (disp_ch >= ch_count) {
					disp_ch = 0;
					sum_period = sum_period == MONTH ? DAY : MONTH;
				}
			}
			c = lcd_pat[disp_ch];

			switch(sum_period){
			case MONTH:
				e = w_est[disp_ch];
				break;
			case DAY:
				c |= 0x04;				// Call icon
				e = w_est[disp_ch] - w_day[disp_ch];
				break;
			default:
				break;
			}
			lcd7seg_PutPattern(c, 9);	// Channel icon

			switch(disp_item) {
			case WH_W:
				if (e < 0) {							// Over flow
					lcd7seg_PutPattern(0b01111001, 8);	// "E"
				} else if (e < 3600000) {				// .999kWh
					lcd7seg_DP(1, 1);
					lcd7seg_PutPattern(0b00000000, 1);	// " "
					lcd7seg_DispInt((e + 1800) / 3600, 2, -3);
				} else if (e < 7200000) {				// 1.999kWh
					lcd7seg_DP(1, 1);
					lcd7seg_DispInt((e + 1800) / 3600, 1, 4);
				} else {								// ZZZ9
					lcd7seg_DP(0, 1);
					lcd7seg_DispInt((e + 1800000) / 3600000, 1, 4);
				}
	
				lcd7seg_DP(0, 7);
				lcd7seg_DispInt(w_val[disp_ch], 5, 4);
				break;
			case W_PEAK:
				lcd7seg_PutPattern(0b00111110, 2);		// "W"
				lcd7seg_PutPattern(0b01000000, 3);		// "-"
				lcd7seg_PutPattern(0b01110011, 4);		// "p"
				lcd7seg_PutPattern(0b00000001, 10);		// overlay column2
				lcd7seg_DP(0, 7);
				lcd7seg_DispInt(w_max[disp_ch], 5, 4);
				break;
			case YEN:
				y = e / 3600 * kwh_price / 1000;
				lcd7seg_PutPattern(0b00000000, 1);		// " "
				lcd7seg_PutPattern(0b01101110, 2);		// "y"
				lcd7seg_PutPattern(0b01111011, 3);		// "e"
				lcd7seg_PutPattern(0b01010100, 4);		// "n"
				if ((y >= 0) && (y < 10000)) {
					lcd7seg_DispInt(y, 5, 4);
				} else {
					lcd7seg_PutPattern(0b01111001, 8);	// "E"
				}
				break;
			case VOL:
				lcd7seg_PutPattern(0b00000000, 1);		// " "
				lcd7seg_PutPattern(0b00100010, 2);		// "V"
				lcd7seg_PutPattern(0b00000010, 10);		// overlay column2
				lcd7seg_PutPattern(0b00000000, 3);		// " "
				lcd7seg_PutPattern(0b00000000, 4);		// " "
				lcd7seg_DispInt(v_val, 5, 4);
				break;
			case AMP:
				lcd7seg_PutPattern(0b00000000, 1);		// " "
				lcd7seg_PutPattern(0b01110111, 3);		// "A"
				lcd7seg_PutPattern(0b00000000, 4);		// " "
	
				if (a_val[disp_ch] <= 999) {			//
					lcd7seg_DP(0, 7);					// 999mA 
					lcd7seg_DispInt(a_val[disp_ch], 5, -4);
					lcd7seg_PutPattern(0b00010100, 2);	// "m"
					lcd7seg_PutPattern(0b00000010, 10);	// overlay column2
				} else {								//
					lcd7seg_DP(1, 7);					// 999.9A
					lcd7seg_DispInt((a_val[disp_ch]+50)/100, 5, 4);
					lcd7seg_PutPattern(0b00000000, 2);	// " "
					lcd7seg_PutPattern(0b00000000, 10);	// " "
				}
				break;
			case PF:
				lcd7seg_PutPattern(0b00000000, 1);		// " "
				lcd7seg_PutPattern(0b01110011, 2);		// "P"
				lcd7seg_PutPattern(0b01110001, 3);		// "F"
				lcd7seg_PutPattern(0b00000000, 4);		// " "
	
				if ((v_val != 0) && (a_val[disp_ch] != 0)) {
					lcd7seg_DispInt(
						(long)w_val[disp_ch] * 1000 * 100 / ((long)v_val * (long)a_val[disp_ch]),
						5, 4);
				} else {
					lcd7seg_DispInt(9999, 5, 4);	// Overflow
				}
				break;
			case DATE_TIME:
				lcd7seg_PutPattern(0x28, 9);		// Colon
				lcd7seg_DispInt(mm, 1, 2);
				lcd7seg_DispInt(dd, 3, 2);
				lcd7seg_DispInt(hh, 5, 2);
				lcd7seg_DispInt((int)mi + 100, 7, 2);
				lcd7seg_DP(0, 1);
				break;
			default:
				break;
			}
		}
//		ts_disp = TMR1;				// DEBUG

		//
		// Wait next
		//
		SLEEP();
	}
}
